import threading
import time

import cv2
import hailo
import numpy as np
from gi.repository import Gst

from hailo_rpi_common import get_caps_from_pad, get_numpy_from_buffer
from pose_estimation_pipeline import GStreamerPoseEstimationApp


# 居眠り検知用のコールバッククラスの定義
class user_app_callback_class:
    def __init__(self):
        # 居眠り検知フラグ
        self.drowsiness_detected = False
        # スレッドをロックし意図しない変更を防ぐ
        self.lock = threading.Lock()
        # 居眠り開始時間
        self.drowsiness_start_time = None

    # 居眠り状態をセットするメソッド
    def set_drowsiness(self, detected):
        with self.lock:
            if detected and not self.drowsiness_detected:
                # 居眠りが新たに検知された場合、開始時間を記録
                self.drowsiness_start_time = time.time()
            elif not detected:
                # 居眠りが解除された場合、開始時間をリセット
                self.drowsiness_start_time = None
            self.drowsiness_detected = detected

    # 居眠り状態をゲットするメソッド
    def get_drowsiness(self):
        with self.lock:
            return self.drowsiness_detected

    # 居眠り状態の時間をゲットするメソッド
    def get_drowsiness_duration(self):
        with self.lock:
            if self.drowsiness_detected and self.drowsiness_start_time:
                return time.time() - self.drowsiness_start_time
            return 0


# コールバック関数を定義
def app_callback(pad, info, user_data):
    GstPadProbeReturn = Gst.PadProbeReturn

    # GstBufferを取得
    buffer = info.get_buffer()
    if buffer is None:
        return GstPadProbeReturn.OK

    # パッドからCaps（画像情報）を取得
    format, width, height = get_caps_from_pad(pad)

    # フレームを取得
    frame = None
    if user_data.use_frame and format and width and height:
        frame = get_numpy_from_buffer(buffer, format, width, height)

    # Hailoから検出情報を取得
    roi = hailo.get_roi_from_buffer(buffer)
    detections = roi.get_objects_typed(hailo.HAILO_DETECTION)

    # キーポイントのマッピング
    keypoints = get_keypoints()

    # pointsを初期化
    points = None

    # 検出されたオブジェクトを解析
    for detection in detections:
        # ラベル、バウンディングボックスを取得
        label = detection.get_label()
        bbox = detection.get_bbox()
        confidence = detection.get_confidence()
        # 人間が検出された場合のみ以下を実行
        if label == "person" and confidence >= 0.2:
            # 人間のキーポイント（landmarks）を取得し、存在する場合に処理を継続
            landmarks = detection.get_objects_typed(hailo.HAILO_LANDMARKS)
            if landmarks:
                points = landmarks[0].get_points()
                # 各キーポイントの位置と確信度を取得
                for part in [
                    'nose', 'left_eye', 'right_eye', 'left_ear',
                    'right_ear', 'left_shoulder', 'right_shoulder',
                    'left_elbow', 'right_elbow', 'left_wrist',
                    'right_wrist', 'left_hip', 'right_hip',
                    'left_knee', 'right_knee', 'left_ankle',
                    'right_ankle'
                ]:
                    keypoint_index = keypoints.get(part)
                    # キーポイントが無効の場合に処理をスキップ
                    if keypoint_index is None or keypoint_index >= len(points):
                        continue
                    point = points[keypoint_index]
                    point_confidence = point.confidence()

                    # キーポイント座標を取得
                    x = int((point.x() * bbox.width() + bbox.xmin()) * width)
                    y = int((point.y() * bbox.height() + bbox.ymin()) * height)

                    # キーポイント位置を描画
                    if user_data.use_frame:
                        cv2.circle(frame, (x, y), 3, (0, 255, 0), -1)

    if user_data.use_frame and frame is not None:
        # フレームをBGRに変換してキューにセット
        frame = cv2.cvtColor(frame, cv2.COLOR_RGB2BGR)
        user_data.set_frame(frame)

    # 居眠り検知ロジック
    try:
        if points is None:
            print("身体のキーポイントが見つかりません．")
            return GstPadProbeReturn.OK

        # COCOのキーポイントインデックスを取得
        COCO_KEYPOINTS = get_keypoints()

        head_indices = [
            COCO_KEYPOINTS['nose'],
            COCO_KEYPOINTS['left_eye'],
            COCO_KEYPOINTS['right_eye'],
            COCO_KEYPOINTS['left_ear'],
            COCO_KEYPOINTS['right_ear']
        ]
        head_y = []

        # 各キーポイントにアクセス
        for idx in head_indices:
            point = points[idx]
            point_confidence = point.confidence()

            # キーポイントの確信度が0.2より大きい場合にy軸の値をhead_y配列に格納
            if point_confidence > 0.2:
                # point.y()はバウンディングボックス内の相対的なキーポイント位置の割合
                # bbox.ymin()は全体画像に対する相対的なバウンディングボックス位置の割合
                # y_coordは、そのキーポイントの画像内の絶対y座標（大きいほど下に位置）
                y_coord = point.y() * height + bbox.ymin() * height
                head_y.append(y_coord)

        if not head_y:
            print("有効な頭の座標が検出されません")
            return GstPadProbeReturn.OK

        # 頭キーポイント群の平均y座標
        avg_head_y = np.mean(head_y)

        # 肩についても頭と同様に処理
        upper_body_indices = [
            COCO_KEYPOINTS['left_shoulder'],
            COCO_KEYPOINTS['right_shoulder']
        ]
        upper_body_y = []
        for idx in upper_body_indices:
            point = points[idx]
            point_confidence = point.confidence()

            if point_confidence > 0.2:
                y_coord = point.y() * height + bbox.ymin() * height
                upper_body_y.append(y_coord)

        if not upper_body_y:
            print("有効な肩の座標が検出されません")
            return GstPadProbeReturn.OK

        # 肩キーポイント群の平均y座標
        avg_upper_body_y = np.mean(upper_body_y)

        # 上半身のy軸の値の平均よりも頭のy軸の値の平均が小さい場合（頭が肩より下にある場合）にアラート
        if avg_head_y >= avg_upper_body_y:
            if not user_data.get_drowsiness():
                print("Alert: 居眠りが検知されました！")
                user_data.set_drowsiness(True)
        else:
            user_data.set_drowsiness(False)

    except Exception as e:
        print(f"Error: {e}")

    return GstPadProbeReturn.OK


# 一定秒数居眠り検知された場合に警告する関数
def monitor_drowsiness(user_data, threshold=5):
    while True:
        if user_data.get_drowsiness():
            duration = user_data.get_drowsiness_duration()
            if duration >= threshold:
                print("*****Alert: 5秒以上居眠りが検知されました！起きてください！*****")
                user_data.set_drowsiness(False)
        time.sleep(1)


# COCOキーポイントの名前とindexの変換表
def get_keypoints():
    keypoints = {
        'nose': 0,
        'left_eye': 1,
        'right_eye': 2,
        'left_ear': 3,
        'right_ear': 4,
        'left_shoulder': 5,
        'right_shoulder': 6,
        'left_elbow': 7,
        'right_elbow': 8,
        'left_wrist': 9,
        'right_wrist': 10,
        'left_hip': 11,
        'right_hip': 12,
        'left_knee': 13,
        'right_knee': 14,
        'left_ankle': 15,
        'right_ankle': 16,
    }
    return keypoints


if __name__ == "__main__":
    # ユーザアプリコールバッククラスのインスタンスを作成
    user_data = user_app_callback_class()

    # 居眠り監視スレッドを開始
    monitor_thread = threading.Thread(target=monitor_drowsiness,
                                      args=(user_data,))
    monitor_thread.daemon = True
    monitor_thread.start()

    # GStreamerPoseEstimationAppのインスタンスを作成し、コールバック関数を設定
    app = GStreamerPoseEstimationApp(app_callback, user_data)

    # パイプラインを実行
    app.run()
